"""
    ~ const ~
    This module contains all the different constants that
    can be common to more modules
"""
C_DAY_TO_SECOND = 86400
C_HOUR_TO_SECOND= 3600
C_MINUTE_TO_SECOND = 60

C_TASK_SLEEP_DELAY   = 0.010  # Advertiser task sleep
C_VC_CONNECT_TIMEOUT = 3    # Timeout of voice channels

C_FILE_NAME_FORBIDDEN_CHAR = ('<','>','"','/','\\','|','?','*',":")
